/*
    Copyright 2004,2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define VER    "0.1.1"
#define BUFFSZ    32768


void len_err(void);
void std_err(void);


struct head {
    unsigned int    sign;
    unsigned int    lumps;
    unsigned int    offset;
} head;

struct lump {
    unsigned int    offset;
    unsigned int    size;
    unsigned char    name[8];
} lump;


int main(int argc, char *argv[]) {
    FILE        *fd,
            *fdout;
    unsigned char    buff[BUFFSZ],
            filename[9],
            *stri;
    unsigned int    lumppos,
            currlen,
            totext = 0;
    int        err,
            len,
            i,
            first,
            mus = 0,
            verbose = 0;

    printf("\n"
        "WAD Extractor %s\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", VER);

    setbuf(stdout, NULL);

    if(argc < 3) {
        printf("\nUsage: %s [options] <file.WAD> <extract_dir>\n"
            "\nOptions:\n"
            "-m        MUS files identification (.MUS ext)\n"
            "-v        verbose: lump/file names and size\n"
            "\n", argv[0]);
        exit(1);
    }


    argc -= 2;
    for(i = 1; i < argc; i++) {
        switch(argv[i][1]) {
            case 'm': mus = 1; break;
            case 'v': verbose = 1; break;
            default: {
                printf("\nError: wrong argument (%s) in command line\n", argv[i]);
                exit(1);
            }
        }
    }


    fd = fopen(argv[argc], "rb");
    if(!fd) std_err();


    /* extract directory */
    err = chdir(argv[argc + 1]);
    if(err < 0) std_err();


    /* READ HEADER */
    err = fread(&head, 1, sizeof(head), fd);
    if(err < sizeof(head)) len_err();

    lumppos = head.offset;
    if(verbose) printf("\nLumps: %u\n\n", head.lumps);

    for(i = 0; i < head.lumps; i++) {
        err = fseek(fd, lumppos, SEEK_SET);
        if(err < 0) std_err();

        err = fread(&lump, 1, sizeof(lump), fd);
        if(err < sizeof(lump)) len_err();

        lumppos = ftell(fd);

        /* bad chars check */
        memcpy(filename, lump.name, 8);
        filename[8] = 0;
        stri = strchr(filename, '\\');
        if(stri) *stri = '.';

        if(verbose) printf("%s: %u\n", filename, lump.size);

        err = fseek(fd, lump.offset, SEEK_SET);
        if(err != 0) std_err();


        fdout = fopen(filename, "wb");
        if(!fdout) std_err();

        if(lump.size < BUFFSZ) {
            len = fread(buff, 1, lump.size, fd);
            if(len < lump.size) len_err();

            err = fwrite(buff, 1, len, fdout);
            if(err < len) std_err();

            /* MUS identification */
            if(mus) {
                if(!memcmp(buff, "MUS", 3)) {
                    sprintf(buff, "%s.mus", filename);
                    printf("MUS: %s\n", buff);
                    err = rename(filename, buff);
                    if(err < 0) std_err();
                }
            }
        } else {
            currlen = BUFFSZ;
            first = 1;
            while(1) {
                len = fread(buff, 1, currlen, fd);
                if(len < currlen) len_err();

                /* MUS identification */
                if(mus) {
                    if(first) {
                        if(!memcmp(buff, "MUS", 3)) {
                            sprintf(buff, "%s.mus", filename);
                            printf("MUS: %s\n", buff);
                            err = rename(filename, buff);
                            if(err < 0) std_err();
                        }
                    }
                    first = 0;
                }

                lump.size -= len;
                if(lump.size == 0) break;
                if(lump.size < currlen) currlen = lump.size;

                err = fwrite(buff, 1, len, fdout);
                if(err < len) std_err();
            }
        }

        fclose(fdout);
        totext++;
    }

    fclose(fd);

    printf("\n\n%u files extracted\n\n", totext);

    return(0);
}


void len_err(void) {
    printf("\nError: File is corrupted\n");
    exit(1);
}


void std_err(void) {
    perror("\nError");
    exit(1);
}

